import json
import os

from sqlalchemy import create_engine
from sqlalchemy.orm import Session, scoped_session, sessionmaker


class ConnectionCredentialError(RuntimeError):
    pass


class Connection:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Connection, cls).__new__(cls)
        return cls._instance

    def __init__(
        self,
        user: str = None,
        endpoint: str = None,
        port: int = None,
        name: str = None,
        password: str = None,
    ) -> None:
        self._initialized = True
        self._session = None
        self._engine = None
        self.user = user or os.environ.get("DB_USER")
        self.endpoint = endpoint or os.environ.get("DB_ENDPOINT")
        self.port = port or os.environ.get("DB_PORT", "5432")
        self.name = name or os.environ.get("DB_NAME")
        self.password = password or os.environ.get("DB_PASSWORD", None)
        self.region = os.environ.get("AWS_REGION")
        self.environment = os.environ.get("ENVIRONMENT", "development")

        if not (
            self.user and self.endpoint and self.name and self.port and self.password
        ):
            raise ConnectionCredentialError("Invalid database credentials.")

    @property
    def session(self) -> Session:
        if self._session:
            return self._session

        self._session = self.create_session()
        return self._session

    def create_session(self) -> Session:
        session_factory = sessionmaker(
            bind=self._get_engine(), future=True, expire_on_commit=False
        )
        session: Session = scoped_session(session_factory)
        return session

    def get_connection_string(self):
        conn = "postgresql+pg8000://{}:{}@{}:{}/{}".format(
            self.user, self.password, self.endpoint, self.port, self.name
        )

        return conn

    def _get_engine(self):
        if self._engine:
            return self._engine

        self._engine = create_engine(self.get_connection_string())

        return self._engine



from __future__ import annotations

from dataclasses import field
from typing import Dict, Union

from marshmallow.utils import EXCLUDE
from marshmallow_dataclass import dataclass
from uuid6 import UUID
from uuid6 import uuid7 as uuid7


class BaseSchema:
    @classmethod
    def load(cls, data: Union[dict, BaseSchema], unknown=EXCLUDE, **kwargs):
        if isinstance(data, BaseSchema):
            data = data.dump()

        return cls.Schema().load(data, unknown=unknown, **kwargs)

    def dump(self, skip_none=False, **kwargs):
        data = self.Schema(**kwargs).dump(self)

        if skip_none:
            data = {key: val for key, val in data.items() if val is not None}

        return data



import sys
from dataclasses import MISSING
from dataclasses import field as datafield
from dataclasses import fields
from datetime import date, datetime
from enum import EnumType
from typing import Any, Dict, List, Type, get_type_hints
from uuid import UUID

from pluralizer import Pluralizer
from sqlalchemy import Column, Index, Table, UniqueConstraint
from sqlalchemy.dialects.postgresql import UUID as SQLAlchemyUUID
from sqlalchemy.ext.indexable import index_property
from sqlalchemy.orm import registry, relationship
from sqlalchemy.types import Boolean, Date, DateTime, Float, Integer, String

from hojo.base import BaseModel


def pluralize(word):
    return Pluralizer().plural(word)


def field(
    primary_key: bool = False,
    index: bool = False,
    unique: bool = False,
    has_many: BaseModel = None,
    belongs_to: BaseModel = None,
    has_one: BaseModel = None,
    **kwargs,
) -> Any:
    custom_metadata = {
        "primary_key": primary_key,
        "index": index,
        "unique": unique,
        "has_many": has_many,
        "belongs_to": belongs_to,
        "has_one": has_one,
    }

    # If metadata already exists in kwargs, update it; otherwise, create it
    if "metadata" in kwargs:
        kwargs["metadata"].update(custom_metadata)
    else:
        kwargs["metadata"] = custom_metadata

    return datafield(**kwargs)


class TypeTranslator:
    def __init__(self, model):
        self.model = model
        self.type_mapping = {
            "int": Integer,
            "float": Float,
            "str": String,
            "bool": Boolean,
            "datetime": DateTime,
            "date": Date,
            "UUID": SQLAlchemyUUID(as_uuid=True),
        }

    def translate(self, field_name):
        type_hints = get_type_hints(self.model)
        field_type = type_hints[field_name]

        field_type_name = field_type.__name__
        if field_type_name in self.type_mapping:
            return self.type_mapping[field_type_name]
        elif isinstance(field_type, EnumType):
            return String
        else:
            raise TypeError(
                f"Unsupported type: {field_type} for {self.model.__name__}.{field_name}'"
            )


class RelationshipManager:
    def __init__(self):
        self.relationships = {}

    def add_relationship(self, field_name, field_info):
        pass

    def get_relationships(self):
        return self.relationships


class TableBuilder:
    def __init__(self, mapper_registry, model):
        self.mapper_registry = mapper_registry
        self.model = model
        self.translator = TypeTranslator(model)
        self.relationship_manager = RelationshipManager()

    def build_table(self):
        columns = []
        uniques = {}
        indexes = {}

        for field_info in fields(self.model):
            if self._is_relationship_field(field_info):
                self.relationship_manager.add_relationship(field_info.name, field_info)
                continue

            column_type = self.translator.translate(field_info.name)

            is_primary_key = field_info.name == "id"  # id is the primary key by default
            is_nullable = field_info.default is None

            is_index = self.get_contraint("index", indexes, field_info)
            is_unique = self.get_contraint("unique", uniques, field_info)

            columns.append(
                Column(
                    field_info.name,
                    column_type,
                    primary_key=is_primary_key,
                    nullable=is_nullable,
                    index=is_index,
                    unique=is_unique,
                )
            )

        for index in indexes:
            idx_fields = sorted(indexes[index])
            index_name = "ix_" + "_".join(idx_fields)
            columns.append(Index(index_name, *idx_fields))

        table = Table(
            pluralize(self.model.__name__.lower()),
            self.mapper_registry.metadata,
            *columns,
            extend_existing=True,
        )

        return table

    def get_contraint(self, contraint_type, contraints, field_info):
        idx = field_info.metadata.get(contraint_type)
        match idx:
            case bool():
                is_index = idx
            case str():
                is_index = False
                contraints[idx] = contraints.get(idx) or []
                contraints[idx].append(field_info.name)
            case _:
                is_index = False
        return is_index

    def _is_relationship_field(self, field_info):
        info = getattr(field_info, "metadata") or {}
        return info.get("has_many") or info.get("belongs_to") or info.get("has_one")


def generate_orm_mapping(mapper_registry, model: Type[BaseModel]) -> BaseModel:
    table_builder = TableBuilder(mapper_registry, model)
    table = table_builder.build_table()

    mapper_registry.map_imperatively(
        model,
        table,
    )

    return model



from __future__ import annotations

from typing import ClassVar, List

from hojo.orm.manager import Manager, ModelDescriptor
from hojo.schema import BaseSchema


class BaseModel(BaseSchema):
    _registry: List[BaseModel] = []

    objects: ClassVar = ModelDescriptor(Manager())

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._registry.append(cls)
